import 'package:knt_bloc/knt_bloc.dart';
import 'package:native_admob_flutter/native_admob_flutter.dart';

import '../repository/monetized_repo.dart';
import '../service/admob_service.dart';

/// Handles generating & loading [BannerAdController]/[NativeAdController].
///
/// These controller will be used to create [BannerAd]/[NativeAd] widget.
class StaticAdsBloc extends BaseBloc {
  final AdmobService _admobService;
  final MonetizedRepo _monetizedRepo;

  BannerAdController? _currentBannerAdController;
  late int _retryLoadBannerAd;

  NativeAdController? _currentNativeAdController;
  late int _retryLoadNativeAd;

  StaticAdsBloc(this._admobService, this._monetizedRepo) {
    watch<BaseEvent>(
      _handleEvent,
      transformer: sequential(),
    );
  }

  @override
  void reset() {
    _currentBannerAdController = null;
    _retryLoadBannerAd = 0;
    _currentNativeAdController = null;
    _retryLoadNativeAd = 0;
  }

  Future<void> _handleEvent(
    BaseEvent event,
    Emitter<BaseState> emit,
  ) async {
    switch (event.runtimeType) {
      case PrepareBannerAdEvent:
        return _handlePrepareBannerAdEvent(event as PrepareBannerAdEvent, emit);
      case RefreshBannerAdEvent:
        return _handleRefreshBannerAdEvent(event as RefreshBannerAdEvent, emit);
      case _LoadDoneBannerAdEvent:
        return _handleLoadDoneBannerAdEvent(
            event as _LoadDoneBannerAdEvent, emit);

      case PrepareNativeAdEvent:
        return _handlePrepareNativeAdEvent(event as PrepareNativeAdEvent, emit);
      case RefreshNativeAdEvent:
        return _handleRefreshNativeAdEvent(event as RefreshNativeAdEvent, emit);
      case _LoadDoneNativeAdEvent:
        return _handleLoadDoneNativeAdEvent(
            event as _LoadDoneNativeAdEvent, emit);
    }
  }

  Future<void> _handlePrepareBannerAdEvent(
    PrepareBannerAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _createAndLoadBannerAd();
  }

  Future<void> _handleRefreshBannerAdEvent(
    RefreshBannerAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _handleRefreshBannerAd(emit);
  }

  Future<void> _handleLoadDoneBannerAdEvent(
    _LoadDoneBannerAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _handleRefreshBannerAd(emit);
  }

  Future<void> _handleRefreshBannerAd(Emitter<BaseState> emit) async {
    final purchasedIap = await _monetizedRepo.fetchPurchasedIapStatus();
    if (purchasedIap) {
      reset();
    }
    emit(LoadedBannerAdState(
      bannerAdController: _currentBannerAdController,
    ));
  }

  Future<void> _handlePrepareNativeAdEvent(
    PrepareNativeAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _createAndLoadNativeAd();
  }

  Future<void> _handleRefreshNativeAdEvent(
    RefreshNativeAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _handleRefreshNativeAd(emit);
  }

  Future<void> _handleLoadDoneNativeAdEvent(
    _LoadDoneNativeAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _handleRefreshNativeAd(emit);
  }

  Future<void> _handleRefreshNativeAd(
    Emitter<BaseState> emit,
  ) async {
    final purchasedIap = await _monetizedRepo.fetchPurchasedIapStatus();
    if (purchasedIap) {
      reset();
    }
    emit(LoadedNativeAdState(
      nativeAdController: _currentNativeAdController,
    ));
  }

  Future<void> _createAndLoadBannerAd({bool forceLoadAndShow = true}) async {
    _currentBannerAdController?.dispose();
    final onLoaded = () {
      _retryLoadBannerAd = 0;
      if (forceLoadAndShow) {
        add(_LoadDoneBannerAdEvent());
      }
    };
    final onLoadFailed = () async {
      if (_retryLoadBannerAd < _admobService.maxRetryReloadAds) {
        _retryLoadBannerAd++;
        await _createAndLoadBannerAd();
      }
    };
    _currentBannerAdController = _admobService.createBannerAdController(
      onLoaded: onLoaded,
      onLoadFailed: onLoadFailed,
    );
    try {
      await _currentBannerAdController!.load();
    } on Exception {
      onLoadFailed();
    }
  }

  Future<void> _createAndLoadNativeAd({bool forceLoadAndShow = true}) async {
    _currentNativeAdController?.dispose();
    final onLoaded = () {
      _retryLoadNativeAd = 0;
      if (forceLoadAndShow) {
        add(_LoadDoneNativeAdEvent());
      }
    };
    final onLoadFailed = () async {
      if (_retryLoadNativeAd < _admobService.maxRetryReloadAds) {
        _retryLoadNativeAd++;
        await _createAndLoadNativeAd();
      }
    };
    _currentNativeAdController = _admobService.createNativeAdController(
      onLoaded: onLoaded,
      onLoadFailed: onLoadFailed,
    );
    try {
      await _currentNativeAdController!.load();
    } on Exception {
      onLoadFailed();
    }
  }
}

// region EVENT
class PrepareBannerAdEvent extends BaseEvent {}

class RefreshBannerAdEvent extends BaseEvent {}

class _LoadDoneBannerAdEvent extends BaseEvent {}

class PrepareNativeAdEvent extends BaseEvent {}

class RefreshNativeAdEvent extends BaseEvent {}

class _LoadDoneNativeAdEvent extends BaseEvent {}
// endregion

// region STATE
class LoadedBannerAdState extends BaseState {
  final BannerAdController? bannerAdController;

  const LoadedBannerAdState({this.bannerAdController});
}

class LoadedNativeAdState extends BaseState {
  final NativeAdController? nativeAdController;

  const LoadedNativeAdState({this.nativeAdController});
}
// endregion
